import getpass
import ConfigParser
import json
from os.path import join, expanduser

from tutum.api import auth
from tutum.api import exceptions
import tutum

from tutumcli import utils


TUTUM_FILE = '.tutum'
AUTH_SECTION = 'auth'
USER_OPTION = "user"
APIKEY_OPTION = 'apikey'


def authenticate():

    username = raw_input("Username: ")
    password = getpass.getpass()
    try:
        api_key = auth.get_apikey(username, password)
        if api_key is not None:
            config = ConfigParser.ConfigParser()
            config.add_section(AUTH_SECTION)
            config.set(AUTH_SECTION, USER_OPTION, username)
            config.set(AUTH_SECTION, APIKEY_OPTION, api_key)
            with open(join(expanduser('~'), TUTUM_FILE), 'w') as cfgfile:
                config.write(cfgfile)
            print "Login succeeded!"
    except exceptions.TutumAuthError:
        print "Wrong username and/or password. Please try to login again"
    except Exception as e:
        print e


def apps(quiet=False, status=None):
    try:
        app_list = tutum.Application.list(state=status)
        headers = ["NAME", "UUID", "STATUS", "IMAGE", "SIZE (#)", "DEPLOYED", "WEB HOSTNAME"]
        data_list = []
        long_uuid_list = []
        if len(app_list) != 0:
            for app in app_list:
                data_list.append([app.unique_name, app.uuid[:8], utils.add_unicode_symbol_to_state(app.state),
                                  app.image_name, "%s (%d)" % (app.container_size, app.current_num_containers),
                                  utils.get_humanize_local_datetime_from_utc_datetime_string(app.deployed_datetime),
                                  app.web_public_dns])
                long_uuid_list.append(app.uuid)
        else:
            data_list.append(["", "", "", "", "", "", ""])
        if quiet:
            for uuid in long_uuid_list:
                print uuid
        else:
            utils.tabulate_result(data_list, headers)
    except Exception as e:
        print e


def application_details(identifiers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            print json.dumps(app_details.get_all_attributes(), indent=2)
        except Exception as e:
            print e


def app_start(identifiers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            result = app_details.start()
            if result:
                print app_details.uuid
        except Exception as e:
            print e


def app_stop(identifiers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            result = app_details.stop()
            if result:
                print app_details.uuid
        except Exception as e:
            print e


def app_terminate(identifiers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            result = app_details.delete()
            if result:
                print app_details.uuid
        except Exception as e:
            print e


def app_logs(identifiers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            print app_details.logs
        except Exception as e:
            print e


def app_scale(identifiers, target_num_containers):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            if target_num_containers:
                app_details.target_num_containers = target_num_containers
                result = app_details.save()
                if result:
                    print app_details.uuid
        except Exception as e:
            print e


def app_alias(identifiers, dns):
    for identifier in identifiers:
        try:
            app_details = utils.fetch_app(identifier)
            if dns is not None:
                app_details.web_public_dns = dns
                result = app_details.save()
                if result:
                    print app_details.uuid
        except Exception as e:
            print e


def app_run(image, name, container_size, target_num_containers, run_command, entrypoint, container_ports,
            container_envvars, linked_to_application, autorestart, autoreplace, autodestroy, roles):
    try:
        ports = utils.parse_ports(container_ports)
        envvars = utils.parse_envvars(container_envvars)
        app = tutum.Application.create(image=image, name=name, container_size=container_size,
                                       target_num_containers=target_num_containers, run_command=run_command,
                                       entrypoint=entrypoint, container_ports=ports,
                                       container_envvars=envvars, linked_to_application=linked_to_application,
                                       autorestart=autorestart, autoreplace=autoreplace, autodestroy=autodestroy, roles=roles)
        result = app.save()
        if result:
            print app.uuid
    except Exception as e:
        print e


def ps(app_identifier, quiet=False, status=None):
    try:
        if app_identifier is None:
            containers = tutum.Container.list(state=status)
        elif utils.is_uuid4(app_identifier):
            containers = tutum.Container.list(application__uuid=app_identifier, state=status)
        else:
            containers = tutum.Container.list(application__name=app_identifier, state=status) + \
                         tutum.Container.list(application__uuid__startswith=app_identifier, state=status)
        headers = ["NAME", "UUID", "STATUS", "IMAGE", "RUN COMMAND", "SIZE", "EXIT CODE", "DEPLOYED", "PORTS"]
        data_list = []
        long_uuid_list = []
        if len(containers) != 0:
            for container in containers:
                ports_string = ""
                for index, port in enumerate(container.container_ports):
                    if port['outer_port'] is not None:
                        ports_string += "%s:%d->" % (container.public_dns, port['outer_port'])
                    ports_string += "%d/%s" % (port['inner_port'], port['protocol'])
                    if index != len(container.container_ports) - 1:
                        ports_string += ", "
                data_list.append([container.unique_name, container.uuid[:8],
                                  utils.add_unicode_symbol_to_state(container.state), container.image_name,
                                  container.run_command, container.container_size, container.exit_code,
                                  utils.get_humanize_local_datetime_from_utc_datetime_string(container.deployed_datetime),
                                  ports_string])
                long_uuid_list.append(container.uuid)
        else:
            data_list.append(["", "", "", "", "", "", "", "", ""])
        if quiet:
            for uuid in long_uuid_list:
                print uuid
        else:
            utils.tabulate_result(data_list, headers)
    except Exception as e:
        print e


def container_inspect(identifiers):
    for identifier in identifiers:
        try:
            container_details = utils.fetch_container(identifier)
            print json.dumps(container_details.get_all_attributes(), indent=2)
        except Exception as e:
            print e


def container_start(identifiers):
    for identifier in identifiers:
        try:
            container_details = utils.fetch_container(identifier)
            result = container_details.start()
            if result:
                print container_details.uuid
        except Exception as e:
            print e


def container_stop(identifiers):
    for identifier in identifiers:
        try:
            container_details = utils.fetch_container(identifier)
            result = container_details.stop()
            if result:
                print container_details.uuid
        except Exception as e:
            print e


def container_terminate(identifiers):
    for identifier in identifiers:
        try:
            container_details = utils.fetch_container(identifier)
            result = container_details.delete()
            if result:
                print container_details.uuid
        except Exception as e:
            print e


def container_logs(identifiers):
    for identifier in identifiers:
        try:
            container_details = utils.fetch_container(identifier)
            print container_details.logs
        except Exception as e:
            print e


def images(quiet=False, jumpstarts=False, linux=False):
    try:
        if jumpstarts:
            image_list = tutum.Image.list(starred=True)
        elif linux:
            image_list = tutum.Image.list(base_image=True)
        else:
            image_list = tutum.Image.list(is_private_image=True)
        headers = ["Name", "Description"]
        data_list = []
        name_list = []
        if len(image_list) != 0:
            for image in image_list:
                data_list.append([image.name, image.description])
                name_list.append(image.name)
        else:
            data_list.append(["", ""])

        if quiet:
            for name in name_list:
                print name
        else:
            utils.tabulate_result(data_list, headers)
    except Exception as e:
        print e


def add_image(repository, username, password, description):
    try:
        image = tutum.Image.create(name=repository, username=username, password=password, description=description)
        result = image.save()
        if result:
            print image.name
    except Exception as e:
        print e


def remove_image(repositories):
    for repository in repositories:
        try:
            image = tutum.Image.fetch(repository)
            result = image.delete()
            if result:
                print repository
        except Exception as e:
            print e


def update_image(repositories, username, password, description):
    for repository in repositories:
        try:
            image = tutum.Image.fetch(repository)
            if username is not None:
                image.username = username
            if password is not None:
                image.password = password
            if description is not None:
                image.description = description
            result = image.save()
            if result:
                print image.name
        except Exception as e:
            print e


import argparse


def add_login_parser(subparsers, parent_parser):
    subparsers.add_parser('login', help='Login into Tutum', description='Login into Tutum', parents=[parent_parser])


def add_apps_parser(subparsers, parent_parser):
    apps_parser = subparsers.add_parser('apps', help='List running applications',
                                        description='List running applications',
                                        parents=[parent_parser])
    apps_parser.add_argument("-q", "--quiet", help="Print only long uuids", action='store_true')
    apps_parser.add_argument("-s", "--status", help="Filter applications by status")

    # App common options
    app_common_parser = argparse.ArgumentParser(add_help=False)
    app_common_parser.add_argument("identifier", help="Application's uuid (either long or short) or name", nargs="+")

    create_app_parser = subparsers.add_parser('run', help='Create and run an application',
                                              description='Create and run an application', parents=[parent_parser])
    create_app_parser.add_argument("image", help="the image used to deploy this application in docker format")
    create_app_parser.add_argument("-n", "--name", help="a human-readable name for the application "
                                                        "(default: image_tag without namespace)")
    create_app_parser.add_argument("-s", "--container_size", help="the size of the application containers "
                                                                  "(default: XS, possible values: XS, S, M, L, XL)")
    create_app_parser.add_argument("-t", "--target_num_containers", help="the number of containers to run "
                                                                         "for this application (default: 1)", type=int)
    create_app_parser.add_argument("-r", "--run_command",
                                   help="the command used to start the application containers "
                                        "(default: as defined in the image)")
    create_app_parser.add_argument("--entrypoint",
                                   help="the command prefix used to start the application containers "
                                        "(default: as defined in the image)")
    create_app_parser.add_argument("-p", "--port",
                                   help="set ports i.e. '80/tcp'"
                                        "(default: as defined in the image)", action='append')
    create_app_parser.add_argument("-e", "--env",
                                   help="set environment variables i.e. 'ENVVAR=foo' "
                                        "(default: as defined in the image, plus any link- or role-generated variables)",
                                   action='append')
    create_app_parser.add_argument("-l", "--link",
                                   help="an application's uuid (either long or short) or name to link this application "
                                        "to, i.e. 80ff1635-2d56-478d-a97f-9b59c720e513'] (default: none)",
                                   action='append')
    create_app_parser.add_argument('--autorestart', help="whether the containers should be restarted if they stop "
                                                         "(default: OFF)", choices=['OFF', 'ON_FAILURE', 'ALWAYS'])
    create_app_parser.add_argument('--autoreplace', help="whether the containers should be replaced with a new one if "
                                                         "they stop (default: OFF)",
                                   choices=['OFF', 'ON_FAILURE', 'ALWAYS'])
    create_app_parser.add_argument('--autodestroy', help="whether the containers should be terminated if "
                                                         "they stop (default: OFF)",
                                   choices=['OFF', 'ON_FAILURE', 'ALWAYS'])
    create_app_parser.add_argument('--role', help="Tutum API roles to grant the application, "
                                                  "i.e. 'global' (default: none, possible values: 'global')",
                                   action='append')

    subparsers.add_parser('inspect', help='Inspect an application', description='Inspect an application',
                          parents=[parent_parser, app_common_parser])
    subparsers.add_parser('start', help='Start an application', description='Start an application',
                                             parents=[parent_parser, app_common_parser])
    subparsers.add_parser('stop', help='Stop an application', description='Stop an application', parents=[parent_parser, app_common_parser])
    subparsers.add_parser('terminate', help='Terminate an application', description='Terminate an application',
                          parents=[parent_parser, app_common_parser])
    subparsers.add_parser('logs', help='Get logs from an application', description='Get logs from an application',
                          parents=[parent_parser, app_common_parser])
    scale_app_parser = subparsers.add_parser('scale', help='Scale an application',
                                             description='Scale an application',
                                             parents=[parent_parser, app_common_parser])
    scale_app_parser.add_argument("target_num_containers",
                                  help="target number of containers to scale this application to",
                                  type=int)

    alias_app_parser = subparsers.add_parser('alias', help="Change application's dns",
                                             description="Change application's dns",
                                             parents=[parent_parser, app_common_parser])
    alias_app_parser.add_argument("dns", help="custom domain to use for this web application")


def add_containers_parser(subparsers, parent_parser):
    containers_parser = subparsers.add_parser('ps', help='List running containers',
                                              description='List running containers', parents=[parent_parser])
    containers_parser.add_argument("-i", "--identifier", help="Application's uuid (either long or short) or name")
    containers_parser.add_argument("-q", "--quiet", help="Print only long uuids", action='store_true')
    containers_parser.add_argument("-s", "--status", help="Filter containers by status")

    # Container common options
    container_common_parser = argparse.ArgumentParser(add_help=False)
    container_common_parser.add_argument("identifier", help="Container's uuid (either long or short) or name", nargs="+")

    subparsers.add_parser('inspect-container', help='Inspect a container', description='Inspect a container',
                          parents=[parent_parser, container_common_parser])
    subparsers.add_parser('start-container', help='Start a container', description='Start a container',
                          parents=[parent_parser,  container_common_parser])
    subparsers.add_parser('stop-container', help='Stop a container', description='Stop a container',
                          parents=[parent_parser, container_common_parser])
    subparsers.add_parser('terminate-container', help='Terminate a container', description='Terminate a container',
                          parents=[parent_parser, container_common_parser])
    subparsers.add_parser('logs-container', help='Get logs from a container', description='Get logs from a container',
                          parents=[parent_parser, container_common_parser])


def add_images_parser(subparsers, parent_parser):
    images_parser = subparsers.add_parser('images', help='List private images', description='List private images',
                                          parents=[parent_parser])
    images_parser.add_argument("-q", "--quiet", help="Print only image names", action='store_true')
    image_list_options = images_parser.add_mutually_exclusive_group()
    image_list_options.add_argument("-j", "--jumpstarts", help="List jumpstart images", action='store_true')
    image_list_options.add_argument("-l", "--linux", help="List linux images", action='store_true')

    add_new_image_parser = subparsers.add_parser('add', help='Add a private image', description='Add a private image',
                                                 parents=[parent_parser])
    add_new_image_parser.add_argument("repository", help="Full image repository, i.e. quay.io/tutum/test-repo")
    add_new_image_parser.add_argument("username", help="Username to authenticate with the registry")
    add_new_image_parser.add_argument("password", help="Username password")
    add_new_image_parser.add_argument("-d", "--description", help="Image description")

    # image common options
    image_common_parser = argparse.ArgumentParser(add_help=False)
    image_common_parser.add_argument("repository", help="Full image repository, i.e. quay.io/tutum/test-repo", nargs="+")

    subparsers.add_parser('remove', help='Remove a private image', description='Remove a private image',
                          parents=[parent_parser, image_common_parser])

    update_image_parser = subparsers.add_parser('update', help='Update a private image',
                                                description='Update a private image',
                                                parents=[parent_parser, image_common_parser])
    update_image_parser.add_argument("-u", "--username", help="New username to authenticate with the registry")
    update_image_parser.add_argument("-p", "--password", help="New username password")
    update_image_parser.add_argument("-d", "--description", help="New image description")


import datetime
import re

from tabulate import tabulate
import tutum
from dateutil import tz
import ago

from tutumcli.exceptions import NonUniqueIdentifier, ObjectNotFound, BadParameter


def tabulate_result(data_list, headers):
    print tabulate(data_list, headers, stralign="left", tablefmt="plain")


def _from_utc_string_to_utc_datetime(utc_datetime_string):
    if not utc_datetime_string:
        return None
    utc_date_object = datetime.datetime.strptime(utc_datetime_string, "%a, %d %b %Y %H:%M:%S +0000")

    return utc_date_object


def get_humanize_local_datetime_from_utc_datetime_string(utc_datetime_string):
    local_now = datetime.datetime.now(tz.tzlocal())
    utc_target_datetime = _from_utc_string_to_utc_datetime(utc_datetime_string)
    if utc_target_datetime:
        local_target_datetime = _from_utc_string_to_utc_datetime(utc_datetime_string).replace(tzinfo=tz.gettz("UTC")).astimezone(tz=tz.tzlocal())
        return ago.human(local_now - local_target_datetime, precision=1)
    return ""


def is_uuid4(identifier):
    uuid4_regexp = re.compile('^[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}', re.I)
    match = uuid4_regexp.match(identifier)
    return bool(match)


def fetch_container(identifier):

    if is_uuid4(identifier):
        return tutum.Container.fetch(identifier)
    else:
        objects_same_identifier = tutum.Container.list(unique_name=identifier) or \
                                  tutum.Container.list(uuid__startswith=identifier)
        if len(objects_same_identifier) == 1:
            return objects_same_identifier[0]
        elif len(objects_same_identifier) == 0:
            raise ObjectNotFound("Cannot find a container with the identifier '%s'" % identifier)
        raise NonUniqueIdentifier("More than one container has the same identifier, please use the long uuid")


def fetch_app(identifier):

    if is_uuid4(identifier):
        return tutum.Application.fetch(identifier)
    else:
        objects_same_identifier = tutum.Application.list(unique_name=identifier) or \
                                  tutum.Application.list(uuid__startswith=identifier)
        if len(objects_same_identifier) == 1:
            return objects_same_identifier[0]
        elif len(objects_same_identifier) == 0:
            raise ObjectNotFound("Cannot find an application with the identifier '%s'" % identifier)
        raise NonUniqueIdentifier("More than one application has the same identifier, please use the long uuid")


def parse_ports(port_list):
    if port_list is not None:
        parsed_ports = []
        for port in port_list:
            parsed_ports.append(_get_port_dict(port))
    else:
        parsed_ports = None
    return parsed_ports


def _get_port_dict(port):
    port_regexp = re.compile('^[0-9]{1,5}/(tcp|udp)$', re.I)
    match = port_regexp.match(port)
    if bool(match):
        port = port.split("/", 1)
        inner_port = int(port[0])
        protocol = port[1].lower()
        return {'protocol': protocol, 'inner_port': inner_port}
    raise BadParameter("Port argument %s does not match with 'port/protocol'. Example: 80/tcp" % port)


def parse_envvars(envvar_list):
    if envvar_list is not None:
        parsed_envvars = []
        for envvar in envvar_list:
            parsed_envvars.append(_is_envvar(envvar))
    else:
        parsed_envvars = None
    return parsed_envvars


def _is_envvar(envvar):
    envvar_regexp = re.compile('^[a-zA-Z_]+[a-zA-Z0-9_]*=[^?!=]+$', re.I)
    match = envvar_regexp.match(envvar)
    if bool(match):
        envvar = envvar.split("=", 1)
        return {'key': envvar[0], 'value': envvar[1]}
    raise BadParameter("Environment Variable argument %s does not match with 'KEY=VALUE'. Example: ENVVAR=foo" % envvar)


def add_unicode_symbol_to_state(state):
    if state in ["Running", "Partly running"]:
        return u"\u25B6 " + state
    elif state in ["Init", "Stopped"]:
        return u"\u25FC " + state
    elif state in ["Starting", "Stopping", "Scaling", "Terminating"]:
        return u"\u2699 " + state
    elif state in ["Start failed", "Stopped with errors"]:
        return u"\u0021 " + state
    elif state == "Terminated":
        return u"\u2718 " + state
    return state


